@@ -28,13 +28,11 @@ def parse_args() -> argparse.Namespace:
     parser.add_argument('--depth', metavar='MAX_DEPTH', type=int, help='set the max path length')
     parser.add_argument('--array-lengths', metavar='NAME1=LENGTH1,NAME2=LENGTH2,...', help='set the length of dynamic-sized arrays including bytes and string (default: loop unrolling bound)')
 
-    parser.add_argument('--use-srcmap', action=argparse.BooleanOptionalAction, default=True, help='use source mappings')
-
-    parser.add_argument('--smt-add',          action=argparse.BooleanOptionalAction, default=True,  help='interpret `+`')
-    parser.add_argument('--smt-sub',          action=argparse.BooleanOptionalAction, default=True,  help='interpret `-`')
-    parser.add_argument('--smt-mul',          action=argparse.BooleanOptionalAction, default=True,  help='interpret `*`')
-    parser.add_argument('--smt-div',          action=argparse.BooleanOptionalAction, default=False, help='interpret `/`')
-    parser.add_argument('--smt-div-by-const', action=argparse.BooleanOptionalAction, default=False, help='interpret division by constant')
+    parser.add_argument('--no-smt-add',          action='store_true', help='do not interpret `+`')
+    parser.add_argument('--no-smt-sub',          action='store_true', help='do not interpret `-`')
+    parser.add_argument('--no-smt-mul',          action='store_true', help='do not interpret `*`')
+    parser.add_argument(   '--smt-div',          action='store_true', help=       'interpret `/`')
+    parser.add_argument(   '--smt-div-by-const', action='store_true', help=       'interpret division by constant')
 
     parser.add_argument('--solver-timeout-branching', metavar='TIMEOUT', type=int, default=1000, help='set timeout (in milliseconds) for solving branching conditions (default: %(default)s)')
     parser.add_argument('--solver-timeout-assertion', metavar='TIMEOUT', type=int, default=60000, help='set timeout (in milliseconds) for solving assertion violation conditions (default: %(default)s)')
@@ -43,13 +41,13 @@ def parse_args() -> argparse.Namespace:
     parser.add_argument('-v', '--verbose', action='count', default=0, help='increase verbosity levels: -v, -vv, -vvv, -vvvv')
     parser.add_argument('--debug', action='store_true', help='run in debug mode')
     parser.add_argument('--log', metavar='LOG_FILE_PATH', help='log individual execution steps in JSON')
-    parser.add_argument('--print-revert', action=argparse.BooleanOptionalAction, default=False, help='print reverting paths in verbose mode')
+    parser.add_argument('--print-revert', action='store_true', help='print reverting paths in verbose mode')
 
     cryticparser.init(parser)
 
     return parser.parse_args()
 
-def add_srcmap(ops: List[Opcode], srcmap: List[str], srcs: Dict):
+def add_srcmap(ops: List[Opcode], srcmap: List[str], srcs: Dict) -> None:
     fpath = {} # file id -> file path
     for src in srcs:
         fpath[srcs[src]['id']] = src
@@ -73,11 +71,74 @@ def add_srcmap(ops: List[Opcode], srcmap: List[str], srcs: Dict):
 
         ops[idx].sm = SrcMap(srctext, jump, mdepth)
 
-def run(
+def setup(
     hexcode: str,
     abi: Dict,
     srcmap: List[str],
     srcs: Dict,
+    args: argparse.Namespace,
+    setup_sig: str,
+    options: Dict
+) -> Exec:
+    # bytecode
+    (ops, code) = decode(hexcode)
+    pgm = ops_to_pgm(ops)
+    add_srcmap(ops, srcmap, srcs)
+
+    # solver
+    solver = SolverFor('QF_AUFBV') # quantifier-free bitvector + array theory; https://smtlib.cs.uiowa.edu/logics.shtml
+    solver.set(timeout=args.solver_timeout_branching)
+
+    # storage
+    storage = {}
+
+    # caller
+    caller = BitVec('msg_sender', 256)
+    solver.add(Extract(255, 160, caller) == BitVecVal(0, 96))
+
+    # this
+    this = BitVec('this_address', 256)
+    solver.add(Extract(255, 160, this) == BitVecVal(0, 96))
+
+    # run setup if any
+
+    sevm = SEVM(options)
+
+    setup_ex = sevm.mk_exec(
+        pgm       = { this: pgm },
+        code      = { this: code },
+        storage   = { this: storage },
+        balance   = { this: con(0) },
+        calldata  = [None] * 4,
+        callvalue = con(0),
+        caller    = caller,
+        this      = this,
+        symbolic  = False,
+        solver    = solver,
+    )
+
+    if setup_sig:
+        wstore(setup_ex.calldata, 0, 4, BitVecVal(setup_sig, 32))
+
+        (setup_exs, setup_steps) = sevm.run(setup_ex)
+
+        if len(setup_exs) != 1:
+            if args.debug: print('\n'.join(map(str, setup_exs)))
+            raise ValueError('multiple paths exist in setUp()')
+
+        setup_ex = setup_exs[0]
+
+        setup_opcode = setup_ex.pgm[setup_ex.this][setup_ex.pc].op[0]
+        if (setup_opcode != 'STOP' and setup_opcode != 'RETURN') or setup_ex.failed: raise ValueError('setUp() failed')
+
+        if args.verbose >= 2:
+            print(setup_ex)
+
+    return setup_ex
+
+def run(
+    setup_ex: Exec,
+    abi: Dict,
     funname: str,
     funsig: str,
     funselector: str,
@@ -85,20 +146,6 @@ def run(
     args: argparse.Namespace,
     options: Dict
 ) -> int:
-    #
-    # bytecode
-    #
-
-    (ops, code) = decode(hexcode)
-    add_srcmap(ops, srcmap, srcs)
-
-    #
-    # solver
-    #
-
-    solver = SolverFor('QF_AUFBV') # quantifier-free bitvector + array theory; https://smtlib.cs.uiowa.edu/logics.shtml
-    solver.set(timeout=args.solver_timeout_branching)
-
     #
     # calldata
     #
@@ -168,16 +215,16 @@ def run(
                     raise ValueError('not feasible')
 
     #
-    # storage
+    # callvalue
     #
 
-    storage = {}
+    callvalue = BitVec('msg_value', 256)
 
     #
     # balance
     #
 
-    orig_balance = BitVec('orig_balance', 256)
+    balance = BitVec('this_balance', 256)
 
     #
     # run
@@ -186,21 +233,46 @@ def run(
     start = timer()
 
     sevm = SEVM(options)
-    (exs, steps) = sevm.execute(
-        ops,
-        code,
-        calldata = cd,
-        storage = storage,
-        solver = solver,
-        balance = orig_balance,
-    )
+
+    solver = SolverFor('QF_AUFBV')
+    solver.set(timeout=args.solver_timeout_branching)
+    solver.add(setup_ex.solver.assertions())
+
+    (exs, steps) = sevm.run(Exec(
+        pgm       = setup_ex.pgm.copy(), # shallow copy
+        code      = setup_ex.code.copy(), # shallow copy
+        storage   = deepcopy(setup_ex.storage),
+        balance   = { setup_ex.this: sevm.arith('ADD', balance, callvalue) },
+        #
+        calldata  = cd,
+        callvalue = callvalue,
+        caller    = setup_ex.caller,
+        this      = setup_ex.this,
+        #
+        pc        = 0,
+        st        = State(),
+        jumpis    = {},
+        output    = None,
+        symbolic  = True,
+        #
+        solver    = solver,
+        path      = deepcopy(setup_ex.path),
+        #
+        log       = deepcopy(setup_ex.log),
+        cnts      = deepcopy(setup_ex.cnts),
+        sha3s     = deepcopy(setup_ex.sha3s),
+        storages  = deepcopy(setup_ex.storages),
+        calls     = deepcopy(setup_ex.calls),
+        failed    = setup_ex.failed,
+        error     = setup_ex.error,
+    ))
 
     # check assertion violations
     normal = 0
     models = []
     stuck = []
     for idx, ex in enumerate(exs):
-        opcode = ex.pgm[ex.pc].op[0]
+        opcode = ex.pgm[ex.this][ex.pc].op[0]
         if opcode == 'STOP' or opcode == 'RETURN':
             if ex.failed:
                 gen_model(args, models, idx, ex)
@@ -225,14 +297,14 @@ def run(
     print(f"{passfail} {funsig} (paths: {normal}/{len(exs)}, time: {end - start:0.2f}s, bounds: [{', '.join(dyn_param_size)}])")
     for model, idx, ex in models:
         if model:
-            print(color_warn('Counterexample: ' + str(model)))
+            print(color_warn('Counterexample: ' + str_model(model, args)))
         else:
             print(color_warn('Counterexample: unknown'))
         if args.verbose >= 1:
             print(f'# {idx+1} / {len(exs)}')
             print(ex)
     for opcode, idx, ex in stuck:
-        print(color_warn('Not supported: ' + opcode))
+        print(color_warn('Not supported: ' + opcode + ' ' + ex.error))
         if args.verbose >= 1:
             print(f'# {idx+1} / {len(exs)}')
             print(ex)
@@ -240,7 +312,7 @@ def run(
     # print post-states
     if args.verbose >= 2:
         for idx, ex in enumerate(exs):
-            if args.print_revert or (ex.pgm[ex.pc].op[0] != 'REVERT' and not ex.failed):
+            if args.print_revert or (ex.pgm[ex.this][ex.pc].op[0] != 'REVERT' and not ex.failed):
                 print(f'# {idx+1} / {len(exs)}')
                 print(ex)
 
@@ -255,7 +327,7 @@ def run(
     else:
         return 1
 
-def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec):
+def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec) -> None:
     res = ex.solver.check()
     if res == sat: model = ex.solver.model()
     if res == unknown:
@@ -265,12 +337,14 @@ def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec):
         res = sol2.check()
         if res == sat: model = sol2.model()
     if res == sat and not is_valid_model(model):
-        query = ex.solver.sexpr()
-        query += '(assert (forall ((x (_ BitVec 256)) (y (_ BitVec 256))) (bvule (evm_div x y) x)))'
-    #   query += '(assert (forall ((x (_ BitVec 256)) (y (_ BitVec 256))) (implies (and (not (= x (_ bv0 256))) (not (= y (_ bv0 256))) (= (evm_div (bvmul x y) x) y)) (= (evm_div (bvmul x y) y) x))))'
-        sol3 = Solver(ctx=Context())
+        ctx = Context()
+        sol3 = Solver(ctx=ctx)
         sol3.set(timeout=args.solver_timeout_assertion)
-        sol3.from_string(query)
+        sol3.from_string(ex.solver.sexpr())
+        x = BitVec('x', 256, ctx)
+        y = BitVec('y', 256, ctx)
+        evm_div = f_div.translate(ctx)
+        sol3.add(ForAll([x, y], ULE(evm_div(x, y), x)))
         res = sol3.check()
         if res == sat: model = sol3.model()
     if res == unknown and args.solver_subprocess:
@@ -300,6 +374,18 @@ def is_valid_model(model) -> bool:
             return False
     return True
 
+def str_model(model, args: argparse.Namespace) -> str:
+    def select(var):
+        name = str(var)
+        if name.startswith('p_'): return True
+        elif args.verbose >= 1:
+            if name.startswith('storage') or name.startswith('msg_') or name.startswith('this_'): return True
+        return False
+    if args.debug:
+        return str(model)
+    else:
+        return '[' + ', '.join(sorted(map(lambda decl: f'{decl} = {model[decl]}', filter(select, model)))) + ']'
+
 def main() -> int:
     #
     # z3 global options
@@ -319,12 +405,11 @@ def main() -> int:
         'verbose': args.verbose,
         'debug': args.debug,
         'log': args.log,
-        'add': args.smt_add,
-        'sub': args.smt_sub,
-        'mul': args.smt_mul,
+        'add': not args.no_smt_add,
+        'sub': not args.no_smt_sub,
+        'mul': not args.no_smt_mul,
         'div': args.smt_div,
         'divByConst': args.smt_div_by_const,
-        'srcmap': args.use_srcmap,
         'timeout': args.solver_timeout_branching,
     }
 
@@ -353,9 +438,6 @@ def main() -> int:
     except InvalidCompilation as e:
         raise ValueError('Parse error', e)
 
-    if len(cryticCompile.compilation_units) > 1: raise ValueError('Multiple compilation units', cryticCompile.compilation_units)
-    compilation_unit = list(cryticCompile.compilation_units.values())[0]
-
     #
     # run
     #
@@ -363,39 +445,44 @@ def main() -> int:
     total_passed = 0
     total_failed = 0
 
-    for filename, contracts_names in compilation_unit.filename_to_contracts.items():
-        source_unit = compilation_unit.source_units[filename]
+    for compilation_id, compilation_unit in cryticCompile.compilation_units.items():
 
-        if args.contract:
-            if args.contract not in contracts_names: continue
-            contracts = [args.contract]
-        else:
-            contracts = list(contracts_names)
+        for filename, contracts_names in compilation_unit.filename_to_contracts.items():
+            source_unit = compilation_unit.source_units[filename]
 
-        for contract in contracts:
-            hexcode = source_unit.bytecodes_runtime[contract]
-            srcmap = source_unit.srcmaps_runtime[contract]
-            srcs = []
-            abi = source_unit.abis[contract]
-            methodIdentifiers = source_unit.hashes(contract)
+            if args.contract:
+                if args.contract not in contracts_names: continue
+                contracts = [args.contract]
+            else:
+                contracts = list(contracts_names)
 
-            funsigs = [funsig for funsig in methodIdentifiers if funsig.startswith(args.function)]
+            for contract in contracts:
+                hexcode = source_unit.bytecodes_runtime[contract]
+                srcmap = source_unit.srcmaps_runtime[contract]
+                srcs = {}
+                abi = source_unit.abis[contract]
+                methodIdentifiers = source_unit.hashes(contract)
 
-            if funsigs:
-                num_passed = 0
-                num_failed = 0
-                print(f'\nRunning {len(funsigs)} tests for {filename.short}:{contract}')
-                for funsig in funsigs:
-                    funselector = methodIdentifiers[funsig]
-                    funname = funsig.split('(')[0]
-                    exitcode = run(hexcode, abi, srcmap, srcs, funname, funsig, funselector, arrlen, args, options)
-                    if exitcode == 0:
-                        num_passed += 1
-                    else:
-                        num_failed += 1
-                print(f'Symbolic test result: {num_passed} passed; {num_failed} failed')
-                total_passed += num_passed
-                total_failed += num_failed
+                funsigs = [funsig for funsig in methodIdentifiers if funsig.startswith(args.function)]
+
+                setup_sig = methodIdentifiers.get('setUp()')
+
+                if funsigs:
+                    num_passed = 0
+                    num_failed = 0
+                    print(f'\nRunning {len(funsigs)} tests for {filename.short}:{contract}')
+                    setup_ex = setup(hexcode, abi, srcmap, srcs, args, setup_sig, options)
+                    for funsig in funsigs:
+                        funselector = methodIdentifiers[funsig]
+                        funname = funsig.split('(')[0]
+                        exitcode = run(setup_ex, abi, funname, funsig, funselector, arrlen, args, options)
+                        if exitcode == 0:
+                            num_passed += 1
+                        else:
+                            num_failed += 1
+                    print(f'Symbolic test result: {num_passed} passed; {num_failed} failed')
+                    total_passed += num_passed
+                    total_failed += num_failed
 
     if (total_passed + total_failed) == 0:
         raise ValueError('No matching tests found', args.contract, args.function)
